/*
 * aconversion
 *
 * Usage: aconversion [-V] -t target [-a algorithm] < input.xml
 *
 */

#include <iostream>
#include <getopt.h>

#include <sax/SaxInterface.h>
#include <sax/ParserException.h>

#include "ConversionHandler.h"

using namespace std;
using namespace sax;
using namespace alib;
using namespace conversions;


void help( void )
{
    cout << endl;
    cout << "aconversion 0.01" << endl;
    cout << "Converts between regular expressions, regular grammars and finite automata." << endl;
    cout << "Usage: aconversion -t FORMALISM [SWITCH...]" << endl << endl;
    cout << "  -t, --target=FORMALISM \t Specifies target formalism. See --list-algorithms for available formalisms." << endl;
    cout << "  -a, --algorithm=ALGORITHM \t Specifies algorithm to use. See --list-algorithms." << endl;
    cout << "  -l, --list-algorithms \t Lists available algorithms." << endl;

    cout << endl;
}

void list_algorithms( void )
{
    cout << "List of formalisms:" << endl;
    cout << "\t For finite automata use 'automata', 'fsm' or 'fa'." << endl;
    cout << "\t For regular expression use 'regexp', 're' or 'regex'." << endl;
    cout << "\t For right regular grammar use 'rg', 'rrg' or 'grammar'." << endl;
    cout << "\t For left regular grammar use 'lrg'." << endl;
    cout << "This list is case-insensitive." << endl;

    cout << endl;

    cout << "List of available algorithms:" << endl;

    cout << endl;
    cout << "FA to RE:" << endl;
    cout << "\t" << "'algebraic' - Uses Brzozowski algebraic method (regular equations)." << endl;
    cout << "\t" << "'elimination' [Default] - Performs state elimination algorithm on NFA." << endl;

    cout << endl;
    cout << "FA to RRG:" << endl;
    cout << "\t" << "[Default] - Converts to right regular grammar." << endl;

    cout << endl;
    cout << "FA to LRG:" << endl;
    cout << "\t" << "[Default] - Converts to left regular grammar." << endl;

    cout << endl;
    cout << "RE to FA:" << endl;
    cout << "\t" << "'brzozowski' - Uses Brzozowski's derivation method (WARNING: May get stuck in infinite loop)." << endl;
    cout << "\t" << "'thompson' - Converts to finite automata using Thompson's NFA Construction algorithm." << endl;
    cout << "\t" << "'glushkov' [Default] - Converts to finite automata using Gluskov's NFA algorithm." << endl;

    cout << endl;
    cout << "RE to RRG:" << endl;
    cout << "\t" << "'brzozowski' - Uses Brzozowski's derivation method (WARNING: May get stuck in infinite loop)." << endl;
    cout << "\t" << "'glushkov' [Default] - Converts to finite automata using Gluskov's NFA algorithm." << endl;

    cout << endl;
    cout << "RG to FA:" << endl;
    cout << "\t" << "[Default] - Converts right or left regular grammar to finite automata." << endl;

    cout << endl;
    cout << "RG to RE:" << endl;
    cout << "\t" << "'algebraic' [Default] - Uses Brzozowski algebraic method (regular equations)." << endl;

    cout << endl;
    cout << "RG to RG:" << endl;
    cout << "\t" << "[Default] - Converts right to left regular grammar or vice versa." << endl;
}

int main(int argc, char* argv[])
{
    static struct option long_options[] = {
      {"target",            required_argument,  NULL, 't'},
      {"algorithm",         required_argument,  NULL, 'a'},
      {"version",           no_argument,        NULL, 'v'},
      {"help",              no_argument,        NULL, 'h'},
      {"list-algorithms",   no_argument,        NULL, 'l'},
      {0, 0, 0, 0}
    };

    int long_index = 0, opt = 0;
    string target, algorithm;

    while( ( opt = getopt_long( argc, argv,"a:t:vhl", long_options, & long_index ) )  != -1 )
    {
        switch( opt )
        {
        case 'a':
            algorithm.assign( optarg, strlen( optarg ) );
            break;
        case 't':
            target.assign( optarg, strlen( optarg ) );
            break;
        case 'l':
            list_algorithms( );
            return 0;
        case 'v':
        case 'h':
        default:
            help( );
            return 0;
        }
    }

  list<Token> tokens;
  string input( istreambuf_iterator<char>( cin ), ( istreambuf_iterator<char>( ) ) );
  SaxInterface::parseMemory( input, tokens );

  try
  {
      ConversionHandler conv( tokens, target, algorithm, cout );
      conv.convert( );
  }
  catch( AlibException & e )
  {
      cout << e.what( ) << endl;
      return 1;
  }

  return 0;
}
