/*
 * FSM.h
 *
 *  Created on: Mar 25, 2013
 *      Author: martin
 */

#ifndef FSM_H_
#define FSM_H_

#include <set>
#include "../Automaton.h"
#include "../State.h"
#include "../../alphabet/Symbol.h"
#include "TransitionFSM.h"

namespace automaton {

using namespace std;
using namespace alphabet;

/**
 * Represents Finite State Machine.
 * Can store nondeterministic finite state machines with epsilon transitions.
 */
class FSM: public Automaton {
protected:
	set<TransitionFSM> transitions;
public:
	~FSM();

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual void removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual void removeInputSymbol(const Symbol& symbol);

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransition(const State& current, const Symbol& input, const State& next);

	/**
	 * Adds transition to the automaton.
	 * @param transition transition to add
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransition(const TransitionFSM& transition);

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	void removeTransition(const TransitionFSM& transition);

	/**
	 * @return automaton transitions
	 */
	const set<TransitionFSM>& getTransitions() const;

	/**
	 * @return automaton transitions from state
	 */
	const set<TransitionFSM> getTransitionsFromState(const State& from) const;

	/**
	 * @return automaton transitions to state
	 */
	const set<TransitionFSM> getTransitionsToState(const State& to) const;

	/**
	 * Determines whether FSM contains epsilon transitions.
	 * @return true when automaton doesn't contain epsilon transitions, false otherwise
	 */
	bool isEpsilonFree() const;

	/**
	 * Determines whether FSM is deterministic.
	 * FSM is deterministic when:
	 * \li \c contains only 1 initial state.
	 * \li \c doesn't contain epsilon transitions.
	 * \li doesn't contain more than one transition with combination (from state, input symbol)
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic() const;

	bool isTotal() const;

	/**
	 * @copydoc Automaton::toXML(ostream&) const
	 */
	virtual void toXML(ostream& out) const;
};

} /* namespace automaton */
#endif /* FSM_H_ */
